#include <algorithm>
#include <chrono>
#include <hip/hip_runtime.h>

#include <iostream>
#include <ratio>
#include <vector>

#define HIP_CHECK(expr)                                                        \
  do {                                                                         \
    hipError_t _error = (expr);                                                \
    if (_error != hipSuccess) {                                                \
      std::cerr << "HIP error " << hipGetErrorString(_error) << "(" << _error  \
                << ") in " << __func__ << " at " << __FILE__ << ":"            \
                << __LINE__ << '\n';                                           \
      exit(1);                                                                 \
    }                                                                          \
  } while (false)

int main(int argc, char **argv) {
  if (argc < 2) {
    std::cerr << "A single arguement is required.\n";
    std::cerr << "Usage " << argv[0] << " <hip-module-file>\n";
    exit(1);
  }

  hipModule_t module;
  HIP_CHECK(hipModuleLoad(&module, argv[1]));
  hipFunction_t function;
  HIP_CHECK(hipModuleGetFunction(&function, module, "VkFFT_main"));

  hipEvent_t start;
  hipEvent_t end;
  HIP_CHECK(hipEventCreate(&start));
  HIP_CHECK(hipEventCreate(&end));

  const dim3 grid_size = {1, 1, 112};
  const dim3 block_size = {57, 16, 1};

  const std::size_t size = 8208 * 1024;

  float *d_input = nullptr;
  float *d_output = nullptr;
  HIP_CHECK(hipMalloc(&d_input, size));
  HIP_CHECK(hipMalloc(&d_output, size));

  void *params[] = {&d_input, &d_output};

  using ms_float = std::chrono::duration<float, std::milli>;
  std::vector<ms_float> durations;
  for (int i = 0; i < 100; ++i) {
    HIP_CHECK(hipEventRecord(start, hipStreamDefault));
    for (int j = 0; j < 100; ++j) {
      HIP_CHECK(hipModuleLaunchKernel(
          function, grid_size.x, grid_size.y, grid_size.z, block_size.x,
          block_size.y, block_size.z, 0, hipStreamDefault, params, nullptr));
    }
    HIP_CHECK(hipEventRecord(end, hipStreamDefault));
    HIP_CHECK(hipEventSynchronize(end));

    float ms;
    HIP_CHECK(hipEventElapsedTime(&ms, start, end));

    durations.push_back(ms_float{ms});
  }
  const auto result = *std::min_element(durations.cbegin(), durations.cend());

  std::cout
      << std::chrono::duration_cast<std::chrono::nanoseconds>(result).count()
      << '\n';

  HIP_CHECK(hipEventDestroy(end));
  HIP_CHECK(hipEventDestroy(start));
  HIP_CHECK(hipModuleUnload(module));
}
